/* poprequest.c 
 * V1.2beta          11-May-1997                               */
/*
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY. 
 *
 * While it may be freely distributed, modified and used, it may not be 
 * sold or republished for profit without the permission of the author.
 *
 * Copyright (c) 1997 by Michael Stenns <stenns@vw.tci.uni-hannover.de>
 *
 * The md5 algorithm used by this program is copyrighted by 
 * RSA Data Security, Inc., see notice in md5c.c for details.
 * 
 */

#ifdef GNUC
#define variant_union union
#define variant_struct struct
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>             /* for va_ stuff */

#include <starlet.h>          /* Declare the function*/
#include <lib$routines.h>
#include <descrip.h>
#include <maildef.h>
#include <ssdef.h>
#include <iodef.h>

#ifdef SOCKETSHR
#include <socketshr.h>
#endif /* SOCKETSHR */

#if defined __GNUC__ && defined __ALPHA
#	include <netdb.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#else
#	include <netdb.h>
#	include <socket.h>
#	include <in.h>
#	include <inet.h>
#endif

#include <unixio.h>


#include "md5.h"
#include "poprequest.h"

typedef struct
{
  short buffer_length;
  short item_code;
  void * buffer_address;
  void * return_length_address;
} itmlst3;

int mail$user_begin (int *context, itmlst3 inlist[], itmlst3 outlist[]);
int mail$user_get_info (int *context, itmlst3 inlist[], itmlst3 outlist[]);
int mail$user_end (int *context, itmlst3 inlist[], itmlst3 outlist[]);


/******************************************************************************
 * main routine
 * return values:
 *      SS$_CONNECFAIL  - Error on creating socket or on bind.
 *      SS$_NOSUCHNODE  - Remote node is unknown.
 *      SS$_UNREACHABLE - Network, host or service unreachable.
 *      SS$_REJECT      - Wrong password.
 *      SS$_LINKDISCON  - Network link has been disconnected.
 *      SS$_NORMAL      - Success.
 *****************************************************************************/
int main (int argv, char * argc[])
{
 char *line ;
 unsigned int remote_port;
 int sockfd;
 int status;
 FILE * outfile = NULL;

 cli_dcl_parse();  /* parse the command line */
 
 if (cli_present("HELP")) usage();

 remote_port = atoi (cli_get_value ("PORT"));
 sockfd = connect_to_host (
    (cli_present("NODE")) ? cli_get_value("NODE"): cli_get_value("HOST"), 
    remote_port, &status);
 if (sockfd < 0) exit (status);

 line = get_response (sockfd, &status);
 if (status) l_puts (line);
 status = login_to_server (sockfd,line);
 if (!status)
 {
   fprintf (stderr,"%%POPREQUEST-F-INVPASS, username or password incorrect\n");
   disconnect_from_server (sockfd, 1);
   exit (SS$_REJECT);
 }
 status = execute_xtnd_option (sockfd);
 status = execute_com_option (sockfd);

 status = disconnect_from_server (sockfd, 5);
 exit (status);
}


/******************************************************************************
 * Redirects the output, if requested
 * Call must be after cli_dcl_parse() !
 * return values:
 *    a pointer to the redirected output stream or stdout
 *****************************************************************************/
FILE * redir_output (void)
{
 FILE * outfile = stdout;

 if (cli_present("OUTPUT"))  /* should the output be redirected? */
 {
   FILE * tmpfile;
   char * line = cli_get_value ("OUTPUT");
   if ((tmpfile = fopen (line, "w")))
     outfile = tmpfile;
   else
     fprintf (stderr, "%%POPREQUEST-F-OPEN can't open output file %s\n",line);
 }

 return outfile;
}


/******************************************************************************
 * execute_pop_commands executes an arbitrary pop  commands
 * no syntax check is done
 * return values:
 *    1
 *****************************************************************************/
int execute_pop_command (int sockfd, char * command)
{
 int status = 0;
 char * line;

 l_printf ("\n%s\n",command);
 sendprintf (sockfd,"%s\r\n",command);
 l_puts (get_response (sockfd, &status));
 if (status == 1)  /* command accepted */
 {
   sendprintf (sockfd,"noop\r\n");
   line = get_response (sockfd, &status);
   if (status != 3) 
   {
     status = 0; /* no multiline response */
   }
   do
   {
     if (!line) line = get_response (sockfd, &status);
     if (!status) /* error */
       status = 0;
     else if (!strcmp(line,".")) /* end multiline sequ. */
     {
       line = get_response (sockfd, &status);
       status = 0;
     }
     else
     {
       if (!strncmp (line,"..",2)) line++; /* "byte stuff" */
       l_puts (line);
     }
     line = NULL;
   } while (status && (status <= 3));
 }
 return (status > 3) ? status : SS$_NORMAL;
}


/******************************************************************************
 * execute_com_option executes arbitrary pop  commands
 * no syntax check is done
 * return values:
 *    status
 *****************************************************************************/
int execute_com_option (int sockfd)
{
 int status = SS$_NORMAL;
 char *line;
 FILE *in_file;

 if (cli_present ("INPUT") && (in_file = fopen (cli_get_value ("INPUT"),"r")))
 {
   char record[512];

   while (fgets(record, sizeof record, in_file) && (status == SS$_NORMAL))
   {
     line = Truncate_Input_Record (record);
     if (*line) status = execute_pop_command (sockfd, line);
   }
   fclose (in_file);
 }

 while ((line = cli_get_value ("COMMANDS")) && (status == SS$_NORMAL))
 {
   status = execute_pop_command (sockfd, line);
 }

 return status;
}

/******************************************************************************
 * execute_xtnd_option executes the xtnd commands specified on the 
 * command line. This is a special feature only for IUPOP3 and derivating
 * POP3 servers.
 * return values:
 *    1
 *****************************************************************************/
int execute_xtnd_option (int sockfd)
{
 char ** xtndcmd;
 int ismultiline = 0;
 int status = SS$_NORMAL;
 char * line;

 while ((xtndcmd=extend_xtnd_arg(cli_get_value ("XTND"))) && xtndcmd[0])
 {
   ismultiline = (*xtndcmd[2] == 't') ? 1 : 0;
   l_printf ("\nxtnd %s %s \n", xtndcmd[0], xtndcmd[1] ? xtndcmd[1] : "");
   sendprintf (sockfd,"xtnd %s %s\r\n",xtndcmd[0],xtndcmd[1]?xtndcmd[1]:"");
   l_puts (get_response (sockfd, &status));
   if ((status == 1) && ismultiline)
   {
     do
     {
       line = get_response (sockfd, &status);
       if ((status < 0) || !strcmp(line,".")) /* end multiline sequ. */
         status = 0;
       else
       {
         if (!strncmp (line,"..",2)) line++; /* "byte stuff" */
         l_puts (line);
       }
     } while (status && (status <= 3));
   } 
 }
 return (status > 3) ? status : SS$_NORMAL;
}


/******************************************************************************
 * login to popserver
 * return values:
 *   1  - success
 *   0  - failure
 *****************************************************************************/
int login_to_server (int sockfd, char *line)
{
 int status = 1;
 void *userptr = NULL, *passptr = NULL, *stampptr = NULL;
 char *user = NULL, *password = NULL, *timestamp = NULL, *digest = NULL;
 char *tmp, *tmp2;
 char buffer [80];

 /*
  * parse line for timestamp (all between "<" and ">")
  */
 tmp = tmp2 = NULL;
 tmp = strchr (line, '<');
 if (tmp) tmp2 = strchr (tmp, '>');
 if (tmp2)
 {
   int length = tmp2 - tmp + 1;
   timestamp = stampptr = malloc (length + 1);
   strncpy (timestamp, tmp, length);
 }
 else
   timestamp = "";

 if (cli_present("USERNAME"))
 {
   tmp = cli_get_value ("USERNAME");
   user = userptr = malloc (strlen(tmp) + 1);
   strcpy (user,tmp);
 }
 else
   user = cuserid(NULL); 

 if (cli_present("PASSWORD"))
 {
   tmp = cli_get_value ("PASSWORD");
   if (*tmp == '*') 
     password = readpw ();
   else if (!strcmp(tmp,"LOCAL") || !strcmp(tmp,"APOP"))
     password = passptr = read_local_password ();
   else
   {
     password = passptr = malloc (strlen(tmp) + 1);
     strcpy (password,tmp);
   }
 }

 /*
  *  server login
  *  first try apop, if failure try user/pass combination
  */
 if (password)
 {
   if (!cli_present("APOP")) /* do not use the apop command */
     status = 2;
   else
   {
     digest = create_digest (timestamp, password);
     sendprintf (sockfd, "apop %s %s\r\n",user,digest);
     l_puts (get_response (sockfd, &status));
   }
   if (status == 2) /* "-ERR" */
   {
     sendprintf (sockfd, "user %s\r\n",user);
     l_puts (get_response (sockfd, &status));
     if (status == 1) /* "+OK" */
     {
       sendprintf (sockfd, "pass %s \r\n",password);
       l_puts (get_response (sockfd, &status));
     }
     if (status != 1) status = 0;
   }
   memset (password, 0, strlen(password));
 }
 else
 {
   fprintf (stderr,"%%POPREQUEST-W-NOPASSW missing password\n");
   status = 0;
 }


 if (userptr) free (userptr);
 if (passptr) free (passptr);
 if (stampptr) free (stampptr);
 return status;
}

/******************************************************************************
 * prints usage message
 * return values:
 *   this function does not return
 *****************************************************************************/
void usage (void)
{
  puts ("\nUsage:");
  puts ("======\n");
  puts ("$POPREQUEST /QUALIFIER \n");
  puts ("Known qualifier are:");
  puts ("/HOST=hostname");
  puts ("/PORT=portnumber");
  puts ("/USERNAME=remote_user_name");
  puts ("/PASSWORD[=password]");
  puts ("/[NO]APOP");
  puts ("/OUTPUT=filespec");
  puts ("/INPUT=filespec");
  puts ("/XTND[=(list)]");
  puts ("/COMMANDS[=(list)]");
  puts ("/HELP\n");
  exit (EXIT_SUCCESS);
}

/******************************************************************************
 * disconnect_from_server sends a "quit" message to the server and waits
 * for the answer.
 * return values:
 *   1 for success
 *   2 if server signals failure
 *   0 else
 *****************************************************************************/
int disconnect_from_server (int sockfd, int timeout)
{
 char *line;
 int status, retval = 0;

 if (sockfd)
 {
   status = sendprintf (sockfd, "quit\r\n");
   if (status)
     line = get_response (sockfd, &status);
   if (line) l_puts (line);
 } 
 return retval;
}


/******************************************************************************
 * get_response reads a message line from a POP server
 * Return values:
 *   1 if line starts with a "+OK"
 *   2 if line starts with a "-ERR"
 *   3 other line
 *   0 failure
 *****************************************************************************/
char * get_response (int sockfd, int *status)
{
 char * line;

 line = get_line (sockfd, status); 

 if (line && (*status == SS$_NORMAL))
 {
   if (!strncmp(line,"+OK", 3))
     *status = 1;
   else if (!strncmp(line, "-ERR", 4))
     *status = 2;
   else 
     *status = 3;
 }
 else
 {
   line = "";
 }
 return line;
}


/******************************************************************************
 * get_line returns one line from the POP server.
 * Return values:
 *   pointer to the line received from the POP server without the
 *   trailing \r\n.
 *   A NULL pointer for timeout or failure.
 *****************************************************************************/
char * get_line (int sockfd, int * status)
{
 static char *line, rec_buffer[4096] = {'\0'};
 static char  *rec_buffer_ptr=rec_buffer, *lineptr=rec_buffer;
 static int rec_buffer_len = 0;
 int i, bytes = 0;
 char *tmpptr;

 *status = SS$_NORMAL;
 while (!(tmpptr = strstr(lineptr,"\r\n")) && (bytes >= 0))
 {
   /* move remaining bytes to the begin of the receive buffer */
   for (i=0,rec_buffer_ptr=rec_buffer;i<rec_buffer_len;i++) 
     *rec_buffer_ptr++ = *lineptr++;
   rec_buffer_ptr = rec_buffer + rec_buffer_len;
   *rec_buffer_ptr = '\0';
   lineptr = rec_buffer;

   bytes = recv (sockfd,rec_buffer_ptr,sizeof rec_buffer-rec_buffer_len-1,0);
   if (bytes > 0) 
   {
     rec_buffer_ptr += bytes;
     rec_buffer_len += bytes;
     *rec_buffer_ptr = '\0';
   }
   else
   {
     fprintf (stderr,"%%POPREQUEST-F-RECDATA,  %s\n",vms_errmsg(errno));
     *status = SS$_LINKDISCON;
     tmpptr = NULL;
     break;
   }
 }
 if (tmpptr)
 {
   bytes = tmpptr - lineptr;
   *tmpptr = '\0';
   line = lineptr;
   lineptr = tmpptr + 2; /* \r\n */
   rec_buffer_len -= bytes + 2;
   while (bytes && iscntrl(line[bytes-1])) {bytes--; line[bytes] = '\0';}
   /* while (*line && iscntrl(*line)) line++; */
 }
 else
   line = NULL;
 return line;
}

/******************************************************************************
 * vms_errmsg converts errno to an readable string
 * return values:
 *   the converted string
 *****************************************************************************/
char * vms_errmsg (int error_number)
{
  char * message;

  message = strerror(errno);
  if (!message) message = "unknown error code";
  return message;
}

/******************************************************************************
 * err_msg returns the error message associated with status code
 * return values:
 *   pointer to the error message
 *****************************************************************************/
char * err_msg (int status)
{
  static char message[256];
  int retlen = 0;

  sys$getmsg (status,&retlen,des(message,sizeof message - 1),0xf,NULL);
  message[retlen] = '\0';
  return message;
}

/******************************************************************************
 * connect_to_host connect to an remote host on the specified port
 * return values:
 *   socket number or -1 on failure
 *****************************************************************************/
int connect_to_host (char *remote_hostname, 
                     unsigned int remote_port, 
                     int *error_code) /* specific error code or SS$_NORMAL on success */
{
 unsigned short int net_port;
 unsigned int net_addr;
 int sockfd;               /* socket file descriptor */
 struct sockaddr local_addr, remote_addr;
 struct hostent *hostinfo;
 int status;

 *error_code = SS$_NORMAL;
 /* create local socket */
 sockfd = socket ( AF_INET, SOCK_STREAM, 0 );
 if (sockfd < 0)
 {
   fprintf (stderr,"%%POPREQUEST-F-SOCKET, error creating socket\n");
   fprintf (stderr,"%%POPREQUEST-F-ERRMSG,  %s\n",vms_errmsg(errno));
   *error_code = SS$_CONNECFAIL;
   return -1;
 }

 /* bind local socket */
 memset (&local_addr, '\0',sizeof local_addr);
 local_addr.sa_family = AF_INET;
#ifdef BIND
 status = bind (sockfd, &local_addr, sizeof local_addr); 
 if (status < 0)
 {
   fprintf (stderr,"%%POPREQUEST-F-NOBIND, error binding socket\n");
   fprintf (stderr,"%%POPREQUEST-F-ERRMSG,  %s\n",vms_errmsg(errno));
   *error_code = SS$_CONNECFAIL;
   return -1;
 }
 else
 {
   struct  sockaddr_in local_in , *plocal_addr = (void *) &local_addr;
   int namelen = sizeof local_in;
   float seconds = 1.3;

   memset (&local_in, '\0',sizeof local_in);
   status = getsockname (sockfd, (struct sockaddr *) &local_in, &namelen);
   fprintf (stderr,"%%POPREQUEST-I-BIND, local port number is %d\n",
               ntohs(local_in.sin_port));
 }
#endif

 /* get host address */
 hostinfo = gethostbyname (remote_hostname);
 if (hostinfo) 
   net_addr = *((unsigned int *)hostinfo->h_addr_list[0]);
 else
   net_addr = inet_addr(remote_hostname);
 if (net_addr == -1)
 {
   fprintf (stderr,"%%POPREQUEST-F-NOHOST, host %s is unknown\n",remote_hostname);
   *error_code = SS$_NOSUCHNODE;
   return -1;
 }
 /*
  * Connect to remote host.
  */
 memset (&remote_addr,0,sizeof remote_addr);
 remote_addr.sa_family = AF_INET;
 net_port = htons ((unsigned short int) remote_port);
 memcpy (remote_addr.sa_data, &net_port, sizeof (unsigned short int));
 memcpy (remote_addr.sa_data + sizeof (unsigned short int), 
         &net_addr, sizeof net_addr);
 status = 1;
 status = connect (sockfd, &remote_addr, sizeof remote_addr);
 if (status < 0)
 {
   fprintf (stderr,"%%POPREQUEST-F-CONNECT, error connecting to remote host\n");
   fprintf (stderr,"%%POPREQUEST-F-ERRMSG,  %s\n",vms_errmsg(errno));
   *error_code = SS$_UNREACHABLE;
   return -1;
 }
 return sockfd;
}


/******************************************************************************
 * extend_xtnd_arg expands an item from the IUPOP3 xtnd command list.
 * return values:
 *****************************************************************************/
char **extend_xtnd_arg (char *item)
{
 static char *retlist[4];
 /* first suboption is the default suboption! */
 static char * xtnd_opt[] = {"show","set",NULL};
 static char * xtnd_log[] = {"info","error","thread","debug",NULL};
 static char * xtnd_kill[] = {NULL};
 static xtnd_options_type xtnd[] = {
     {"stats", NULL, TRUE},
     {"netstats", NULL, TRUE},
     {"options", xtnd_opt, TRUE},
     {"loglevel", xtnd_log, FALSE},
     {"shutdown", NULL, FALSE},
     {"restart", NULL, FALSE},
     {"info", NULL, TRUE},
     {"help", NULL, TRUE},
     {"kill", xtnd_kill, FALSE},
     {NULL,NULL,0}};
 int iopt=0, isub;

 retlist[0] = lower(item);
 retlist[1] = ParseConfigLine (retlist[0]);
 retlist[2] = "false";
 retlist[3] = NULL;

 if (retlist[0])
 {
   for (iopt=0;
        xtnd[iopt].option && 
        strncmp(xtnd[iopt].option,retlist[0],strlen(retlist[0]));
        iopt++);
   retlist[0] = xtnd[iopt].option;
   if (!retlist[0]) retlist[1] = NULL;
   if (xtnd[iopt].ismultiline) retlist[2] = "true";
 }
 
 if (xtnd[iopt].suboption)
 {
   if (retlist[1] && xtnd[iopt].suboption[0])
   {
     for (isub=0;
          xtnd[iopt].suboption[isub] && 
          strncmp(xtnd[iopt].suboption[isub],retlist[1],strlen(retlist[1]));
          isub++);
     retlist[1] = xtnd[iopt].suboption[isub];
   }
   if (!retlist[1]) retlist[1] = xtnd[iopt].suboption[0];
   if (!strcmp(retlist[0],"options") && !strcmp(retlist[1],"set") )
     retlist[2] = "false";
 }

 return retlist;
}


/******************************************************************************
 * read_local_password reads the shared secret from the file POP_SECRET.DAT
 * in the local mail directory. This password is commonly used for the
 * POP3 APOP command.
 *****************************************************************************/
char * read_local_password (void)
{
 static char * password = NULL;
 char path[256];
 char shared_secret [1024];
 int pathlen = 0;
 FILE * secret_file;
 itmlst3 nulllist[] = {{0,0,NULL,NULL}};
 itmlst3 inlist[]  = {{0,0,NULL,NULL},{0,0,NULL,NULL}};
 itmlst3 outlist[]  = {{0,0,NULL,NULL},{0,0,NULL,NULL}};
 int context = 0; 
 int status;

 inlist[0].item_code             = MAIL$_USER_USERNAME;
 inlist[0].buffer_address        = cuserid(NULL);
 inlist[0].buffer_length         = strlen (inlist[0].buffer_address);
 inlist[0].return_length_address = 0;

 outlist[0].item_code             = MAIL$_USER_FULL_DIRECTORY;
 outlist[0].buffer_length         = sizeof path - 1;
 outlist[0].buffer_address        = path;
 outlist[0].return_length_address = &pathlen;

 /* Get a mail user context */
 status = mail$user_begin (&context,nulllist,nulllist);
 if (status & 1) status = mail$user_get_info (&context,inlist,outlist);
 if (status & 1) status = mail$user_end (&context,nulllist,nulllist);
 if (!(status & 1))
   fprintf (stderr,"%%POPREQUEST-W-NOMAILDIR error getting mail directory: %s\n",err_msg(status));
 else if (pathlen < (sizeof path - sizeof "POP_SECRET.DAT"))
 {
   strcpy (&path[pathlen],"POP_SECRET.DAT");
   secret_file = fopen (path, "r", "shr=get");
   if (secret_file)
   {
     fgets (shared_secret, sizeof shared_secret, secret_file);
     fclose (secret_file);
     pathlen = strlen (shared_secret);
     while (pathlen && iscntrl(shared_secret[pathlen]) )
       shared_secret[pathlen--] = '\0';
     if (pathlen)
     {
       password = malloc (++pathlen * sizeof (char));
       strcpy (password, shared_secret);
     }
   }
   else
     fprintf (stderr, "%%POPREQUEST-F-OPEN cannot open shared secret file\n");
 }
 return password;
}


/***************************************************************************
 * ParseConfigLine parses a line with the syntax "Option = something".
 * There should be no whitespace characters at the beginning of Option!
 * If the syntax is ok the line will be separated in Option and something 
 * part and a pointer to "something" is returned. If the syntax check 
 * failed, a NULL pointer is returned.
 ***************************************************************************/
char * ParseConfigLine (char * pszLine)
{
char * pszSomething = NULL;
char * pszRight;

 if (pszLine && (pszLine[0] != '#'))
 {                                                /* exists and no comment */
   pszRight = strpbrk (pszLine, "=:"); /* look for the separator */
   if (pszRight)    /* separator found */
   {
     *pszRight = '\0'; pszRight++;
     pszRight = Truncate_String(pszRight);
     pszLine  = Truncate_String(pszLine);
     if (strlen(pszRight) && strlen(pszLine)) /* now seem's all ok */
     {
       pszSomething = pszRight;
     }
   }
 }
 return pszSomething;
}

/**************************************************************************
 * Truncate_Config_File_Record truncates a record from the configuration
 * file. Comments and leading or trailing whitespace characters are removed.
 *************************************************************************/
char * Truncate_Input_Record (char * Record)
{
char * Temp = NULL;

 if (Record)
 {
   Record = Truncate_String (Record);
   switch (*Record)
   {
      case '#' :
      case '!' :
        *Record = '\0';
        break;
      default :
        break;
   }
 }
 else
   Record = "";
 return Record;
}

/**************************************************************************
 * Truncate_String removes leading or trailing whitespace characters
 * from a string.
 *************************************************************************/
char * Truncate_String (char * pszString)
{
int StringLenght, i;

 if (pszString)
 {
   if ((StringLenght = strlen (pszString)))
   {
     for (i = StringLenght-1; i && isspace(pszString[i]); i--) 
     {
       pszString[i] = '\0';
     }
   }
   if ((StringLenght = strlen (pszString)))
   {
     for (i = 0; (i < StringLenght) && isspace(pszString[i]); i++); 
     pszString = &pszString[i];
   }
 }
 return pszString;
}

/**************************************************************************
 * lower 
 *************************************************************************/
char * lower (char * str)
{
char * tmp;

 /* convert to lower */
 if (str)
 {
    for (tmp = str; *tmp; tmp++) *tmp = tolower (*tmp);
 }
 return str;
}


/**************************************************************************
 * create_digest creates the digest string from timestamp and shared secret
 *************************************************************************/
char * create_digest (char * timestamp, char * secret)
{
  static char digest_hexstring [34];
  MD5_CTX context;
  unsigned char digest[16];
  char *tmp;
  int i;
  int status = 0;  /* 0 means wrong password */

  memset (digest_hexstring, '0', 32);
  digest_hexstring[32] = '\0';
  if (timestamp && secret)
  {
    MD5Init (&context);
    MD5Update (&context,(unsigned char *)timestamp, strlen (timestamp));
    MD5Update (&context,(unsigned char *)secret, strlen (secret));
    MD5Final (digest, &context);
    for (i=0; i<16; i++) sprintf (&digest_hexstring[2*i],"%02.2x",digest[i]);
  }
  return digest_hexstring;
}
 
/**************************************************************************
 * sendprintf sends a formatted line to socket sockfd
 *************************************************************************/
int sendprintf (int sockfd, char *format, ...)
{
  va_list       v_args;
  static char   message[256];
  int           bytes = 0;

  va_start (v_args, format);
  vsprintf (message, format, v_args);
  va_end   (v_args);

  bytes = send (sockfd, message, strlen(message), 0);
  return bytes;
}


/**************************************************************************
 * l_puts sends a line to the current output device
 *************************************************************************/
int l_puts (char * line)
{
  return l_printf ("%s\n", line);
}

/**************************************************************************
 * localprintf sends a formatted line to the current output device
 *************************************************************************/
int l_printf (char *format, ...)
{
  va_list       v_args;
  static FILE   *outfile = NULL;
  int           bytes = 0;

  if (!outfile)  outfile = redir_output();

  va_start (v_args, format);
  bytes = vfprintf (outfile, format, v_args);
  va_end   (v_args);

  return bytes;
}



/**************************************************************************
 * Signal functions 
 *************************************************************************/
char *readpw (void)
{
  static char password[64];
  char cMyChar = ' ', Backstr[] = {8,32,8,0};
  int pw_len = 0, status;
  io_statblk status_block;
  short in_chan = 0;
  short out_chan = 0;

  sys$assign (desz("SYS$INPUT"), &in_chan,0,0);
  sys$assign (desz("SYS$OUTPUT"), &out_chan,0,0);
  /*  Queue the I/O.  */
  printf ("\nEnter password>"); fflush(stdout);
  while ( !iscntrl(cMyChar) && (pw_len < (sizeof password / sizeof (char) - 1)))
  {
    if ( status = (sys$qiow (0, in_chan, IO$_READVBLK|IO$M_NOFILTR|IO$M_NOECHO, 
                       &status_block,0,0,
                       &cMyChar,1,0,0,0,0)) != SS$_NORMAL)
    {
      l_puts (vms_errmsg (status_block.cond_value));
      cMyChar = '\r';
    }
    else
    {
      if ((cMyChar == 127) || (cMyChar == 8))
      {
        cMyChar = ' ';
        password[pw_len] = '\0';
        if (pw_len)
        {
          pw_len--;
          sys$qiow (0, out_chan, IO$_WRITEVBLK,
                   &status_block,0,0,
                   Backstr,3,0,0,0,0);
        }
      }
      else
      {
        password[pw_len++] = cMyChar;
        sys$qiow (0, out_chan, IO$_WRITEVBLK,
                 &status_block,0,0,
                 "*",1,0,0,0,0);
      }
    }
  }
  password[--pw_len] = '\0';  
  sys$dassgn (in_chan);
  sys$dassgn (out_chan);
  printf ("\n");
  return password;
}
